from cvxopt import matrix, spmatrix, sparse
from cvxopt import mul, div, sin, cos, exp

from ..consts import Fx0, Fy0, Gx0, Gy0  # NOQA
from ..consts import Fx, Fy, Gx, Gy  # NOQA
from ..consts import pi
from .base import ModelBase

try:
    from cvxoptklu.klu import linsolve
except ImportError:
    from cvxopt.umfpack import linsolve

from ..utils.math import zeros, ones, agtb, ageb, altb, aleb, aandb, aneb
from ..utils.math import mfloor, mround, mmax, mmin, not0

import logging
logger = logging.getLogger(__name__)


class MPPT(object):
    """MPPT control algorithm"""

    def __init__(self, system, name):
        self._algebs.extend(['pwa'])
        self._fnamey.extend(['P_w^{opt}'])

    def init1(self, dae):
        dae.y[self.pwa] = mmax(mmin(2 * dae.x[self.omega_m] - 1, 1), 0)

    def gcall(self, dae):
        dae.g[self.pwa] = mmax(mmin(2 * dae.x[self.omega_m] - 1, 1),
                               0) - dae.y[self.pwa]
        dae.hard_limit(self.pwa, 0, 1)

    def gycall(self, dae):
        pass

    def jac0(self, dae):
        dae.add_jac(Gy0, -1, self.pwa, self.pwa)
        dae.add_jac(Gx0, 2, self.pwa, self.omega_m)
        dae.add_jac(Gy0, 1e-6, self.pwa, self.pwa)


class Turbine(object):
    """Generic wind turbine model"""

    def __init__(self, system, name):
        self._algebs.extend(['pw', 'cp', 'lamb', 'ilamb'])
        self._fnamey.extend([
            'P_w',
            'c_p',
            '\\lambda',
            '\\frac{1}{\\lambda}',
        ])
        self._states.extend(['theta_p'])
        self._fnamex.extend(['\\theta_p'])
        self._mandatory.extend(['wind'])
        self._params.extend(
            ['Kp', 'nblade', 'ngen', 'npole', 'R', 'Tp', 'ngb', 'H'])
        self._powers.extend(['H'])
        self._times.extend(['Tp'])
        self._data.update({
            'R': 35,
            'ngb': 0.011235,
            'npole': 4,
            'wind': 0,
            'Kp': 10,
            'H': 2,
            'Tp': 3.0,
            'nblade': 3,
            'ngen': 40,
        })
        self._descr.update({
            'R': 'Rotor radius',
            'npole': 'Number of poles',
            'wind': 'Wind time series idx',
            'ngb': 'Gear box ratio',
            'Kp': 'Pitch control gain',
            'H': 'Machine rotor and turbine inertia constant',
            'Tp': 'Pitch control time constant',
            'nblade': 'Number of blades',
            'ngen': 'Number of generators',
        })
        self._units.update({
            'R': 'm',
            'Kp': 'pu',
            'H': 'MWs/MVA',
            'Tp': 's',
        })

    @property
    def phi(self):
        deg1 = pi / 180
        dae = self.system.dae
        above = agtb(dae.x[self.omega_m], 1)
        phi_degree_step = mfloor((dae.x[self.omega_m] - 1) / deg1) * deg1
        return mul(phi_degree_step, above)

    def servcall(self, dae):
        self.copy_data_ext('Wind', 'vw', 'vw', self.wind)
        self.copy_data_ext('Wind', 'rho', 'rho', self.wind)
        self.copy_data_ext('Wind', 'Vwn', 'Vwn', self.wind)
        self.mva_mega = 100e6

    def windpower(self, ngen, rho, vw, Ar, R, omega, theta, derivative=False):
        mva_mega = self.system.mva * 1e6
        lamb = omega * R / vw
        ilamb = 1 / (1 / (lamb + 0.08 * theta) - 0.035 / (theta**3 + 1))
        cp = 0.22 * (116 / ilamb - 0.4 * theta - 5) * exp(-12.5 / ilamb)
        pw = 0.5 * ngen * rho * cp * Ar * vw**3 / mva_mega

        a1 = exp(-12.5 / ilamb)
        a2 = (lamb + 0.08 * theta)**2
        a3 = 116. / ilamb - 0.4 * theta - 5
        a4 = -9.28 / (lamb + 0.08 * theta) ** 2 + \
            12.180 * theta * theta / (theta ** 3 + 1) ** 2 - 0.4
        a5 = 1.000 / (lamb + 0.08 * theta) ** 2 - \
            1.3125 * theta * theta / (theta ** 3 + 1) ** 2

        jac = ones(1, 3)
        jac[0] = ngen * R * a1 * rho * vw * vw * Ar * (
            -12.760 + 1.3750 * a3) / a2 / mva_mega
        jac[1] = ngen * (omega * R * (12.760 - 1.3750 * a3) / a2 +
                         0.330 * a3 * vw) * vw * Ar * rho * a1 / mva_mega
        jac[2] = ngen * 0.110 * rho * (
            a4 + a3 * a5) * a1 * Ar * vw**3 / mva_mega

        return pw, jac

    def init1(self, dae):
        # electrical torque in pu
        # te = mul(
        #     xmu,
        #     mul(dae.x[self.irq], dae.y[self.isd]) - mul(
        #         dae.x[self.ird], dae.y[self.isq]))

        for i in range(self.n):
            if self.te0[i] < 0:
                logger.error('Pe < 0 on bus <{}>. Wind speed initialize failed.'
                             .format(self.bus[i]))

        # wind power in pu
        omega = dae.x[self.omega_m]
        theta = dae.x[self.theta_p]
        pw = mul(self.te0, dae.x[self.omega_m])
        dae.y[self.pw] = pw

        # wind speed initialization loop

        R = 4 * pi * self.system.config.freq * mul(self.R, self.ngb,
                                                   div(1, self.npole))
        AA = pi * self.R**2
        vw = 0.9 * self.Vwn

        for i in range(self.n):
            mis = 1
            iter = 0
            while abs(mis) > self.system.tds.config.tol:
                if iter > 50:
                    self.message(
                        'Wind <{}> init failed. '
                        'Try increasing the nominal wind speed.'.format(
                            self.wind[i]))
                    break

                pw_iter, jac = self.windpower(self.ngen[i], self.rho[i], vw[i],
                                              AA[i], R[i], omega[i], theta[i])

                mis = pw_iter - pw[i]
                inc = -mis / jac[1]
                vw[i] += inc
                iter += 1

        # set wind speed
        dae.x[self.vw] = div(vw, self.Vwn)

        lamb = div(omega, vw, div(1, R))
        ilamb = div(1, lamb + 0.08 * theta) - div(0.035, theta**3 + 1)
        cp = 0.22 * mul(
            mul(116, ilamb) - 0.4 * theta - 5, exp(mul(-12.5, ilamb)))

        dae.y[self.lamb] = lamb
        dae.y[self.ilamb] = ilamb
        dae.y[self.cp] = cp

    def gcall(self, dae):
        dae.g[self.pw] = -dae.y[self.pw] + mul(
            0.5, dae.y[self.cp], self.ngen, pi, self.rho, (self.R)**2,
            (self.Vwn)**3, div(1, self.mva_mega), (dae.x[self.vw])**3)
        dae.g[self.lamb] = -dae.y[self.lamb] + mul(
            4, self.R, self.fn, self.ngb, dae.x[self.omega_m], pi,
            div(1, self.Vwn), div(1, self.npole), div(1, dae.x[self.vw]))
        dae.g[self.cp] = -dae.y[self.cp] + mul(
            -1.1 + mul(25.52, dae.y[self.ilamb]) + mul(-0.08800000000000001,
                                                       dae.x[self.theta_p]),
            exp(mul(-12.5, dae.y[self.ilamb])))
        dae.g[self.ilamb] = div(
            1, dae.y[self.lamb] + mul(
                0.08, dae.x[self.theta_p])) - dae.y[self.ilamb] + mul(
                    -0.035, div(1, 1 + (dae.x[self.theta_p])**3))

    def fcall(self, dae):
        dae.f[self.theta_p] = mul(
            div(1, self.Tp), -dae.x[self.theta_p] + mul(
                self.Kp, self.phi, -1 + dae.x[self.omega_m]))
        dae.anti_windup(self.theta_p, 0, pi)

    def gycall(self, dae):
        dae.add_jac(
            Gy,
            mul(0.5, self.ngen, pi, self.rho, (self.R)**2, (self.Vwn)**3,
                div(1, self.mva_mega), (dae.x[self.vw])**3), self.pw, self.cp)
        dae.add_jac(
            Gy,
            mul(25.52, exp(mul(-12.5, dae.y[self.ilamb]))) + mul(
                -12.5, -1.1 + mul(25.52, dae.y[self.ilamb]) + mul(
                    -0.08800000000000001, dae.x[self.theta_p]),
                exp(mul(-12.5, dae.y[self.ilamb]))), self.cp, self.ilamb)
        dae.add_jac(Gy,
                    -(dae.y[self.lamb] + mul(0.08, dae.x[self.theta_p]))**-2,
                    self.ilamb, self.lamb)

    def fxcall(self, dae):
        dae.add_jac(
            Gx,
            mul(1.5, dae.y[self.cp],
                self.ngen, pi, self.rho, (self.R)**2, (self.Vwn)**3,
                div(1, self.mva_mega), (dae.x[self.vw])**2), self.pw, self.vw)
        dae.add_jac(Gx, mul(-0.088, exp(mul(-12.5, dae.y[self.ilamb]))),
                    self.cp, self.theta_p)
        dae.add_jac(
            Gx,
            mul(-4, self.R, self.fn, self.ngb, dae.x[self.omega_m], pi,
                div(1, self.Vwn), div(1, self.npole), (dae.x[self.vw])**-2),
            self.lamb, self.vw)
        dae.add_jac(
            Gx,
            mul(4, self.R, self.fn, self.ngb, pi, div(1, self.Vwn),
                div(1, self.npole), div(1, dae.x[self.vw])), self.lamb,
            self.omega_m)
        dae.add_jac(
            Gx,
            mul(-0.08,
                (dae.y[self.lamb] + mul(0.08, dae.x[self.theta_p]))**-2) + mul(
                    0.10500000000000001, (dae.x[self.theta_p])**2,
                    (1 + (dae.x[self.theta_p])**3)**-2), self.ilamb,
            self.theta_p)

    def jac0(self, dae):
        dae.add_jac(Gy0, -1, self.pw, self.pw)
        dae.add_jac(Gy0, -1, self.cp, self.cp)
        dae.add_jac(Gy0, -1, self.lamb, self.lamb)
        dae.add_jac(Gy0, -1, self.ilamb, self.ilamb)
        dae.add_jac(Gy0, 1e-6, self.pw, self.pw)
        dae.add_jac(Gy0, 1e-6, self.cp, self.cp)
        dae.add_jac(Gy0, 1e-6, self.lamb, self.lamb)
        dae.add_jac(Gy0, 1e-6, self.ilamb, self.ilamb)
        dae.add_jac(Fx0, -div(1, self.Tp), self.theta_p, self.theta_p)
        dae.add_jac(Fx0, mul(self.Kp, self.phi, div(1, self.Tp)), self.theta_p,
                    self.omega_m)


class WTG4DC(ModelBase, Turbine, MPPT):
    """Wind turbine type IV DC output"""

    def __init__(self, system, name):
        ModelBase.__init__(self, system, name)
        Turbine.__init__(self, system, name)
        MPPT.__init__(self, system, name)
        self._group = 'WTG'
        self._name = 'WTG4DC'
        self._algebs.extend(['isd', 'vsd', 'vsq', 'ps', 'te'])
        self._fnamex.extend(['\\omega_m', 'I_{s, q}'])
        self._fnamey.extend(
            ['I_{s, d}', 'V_{s, d}', 'V_{s, q}', 'P_s', '\\tau_e'])
        self._mandatory.extend(['node1', 'node2', 'dcgen', 'wind'])
        self._params.extend([
            'fn', 'rs', 'xd', 'xq', 'psip', 'Tep', 'Teq', 'pmax', 'pmin',
            'qmax', 'qmin', 'u', 'Sn', 'Kdc', 'Ki', 'Kcoi'
        ])
        self._powers.extend(['H', 'pmax', 'pmin', 'qmax', 'qmin'])
        self._service.extend(['qs0'])
        self._states.extend(['omega_m', 'isq'])
        self._times.extend(['Tep', 'Teq', 'Tp'])
        self._r.extend(['rs', 'xd', 'xq'])
        self._data.update({
            'fn': 60,
            'Sn': 40,
            'Tep': 0.01,
            'Teq': 0.01,
            'dcgen': 0,
            'node1': 0,
            'node2': 0,
            'pmax': 1.0,
            'pmin': 0,
            'psip': 0.1,
            'qmax': 0.6,
            'qmin': -0.6,
            'rs': 0.01,
            'u': 1,
            'wind': 0,
            'xd': 0.08,
            'xq': 0.1,
            'Kdc': 0,
            'Ki': 0,
            'Kcoi': 0,
            'busfreq': None,
            'coi': None,
        })
        self._descr.update({
            'Sn':
            'Power rating',
            'Tep':
            'Active power time constant',
            'Teq':
            'Reactive power time constant',
            'dcgen':
            'Static generator index',
            'node1':
            'DC node 1',
            'node2':
            'DC node 2',
            'pmax':
            'Maximum active power',
            'pmin':
            'Minimum reactive power',
            'psip':
            'Permanent field flux',
            'qmax':
            'Maximum reactive power',
            'qmin':
            'Minimum reactive power',
            'rs':
            'Stator resistance',
            'u':
            'Connection status',
            'wind':
            'Wind time series index',
            'xd':
            'd-axis reactance',
            'xq':
            'q-axis reactance',
            'Kdc':
            'DC voltage droop on P reference',
            'Ki':
            'Bus frequency derivative droop on P reference',
            'Kcoi':
            'COI frequency derivative droop on P reference',
        })
        self._units.update({
            'Sn': 'MVA',
            'Tep': 's',
            'Teq': 's',
            'pmax': 'pu',
            'psip': 'Weber',
            'qmax': 'pu',
            'rs': 'pu',
            'xd': 'pu',
            'xq': 'pu'
        })
        self.calls.update({
            'fcall': True,
            'fxcall': True,
            'gcall': True,
            'gycall': True,
            'init1': True,
            'jac0': True
        })
        self._dc = {
            'node1': 'v1',
            'node2': 'v2',
        }

        self._init()

    def _init(self):
        super(WTG4DC, self)._init()

    def servcall(self, dae):
        self.copy_data_ext('DCgen', 'u', 'u0', self.dcgen)
        self.copy_data_ext('DCgen', 'P', 'p0', self.dcgen)
        self.copy_data_ext('Wind', 'vw', 'vw', self.wind)
        self.copy_data_ext('Wind', 'rho', 'rho', self.wind)
        self.copy_data_ext('Wind', 'Vwn', 'Vwn', self.wind)
        # self.copy_data_ext('Node', 'v', 'v1', self.node1)
        # self.copy_data_ext('Node', 'v', 'v2', self.node2)
        # self.qs0 = 0
        # TODO: Fix this dirty hard code
        if self.busfreq[0] is not None:
            self.copy_data_ext('BusFreq', 'dwdt', 'dwdt', self.busfreq)
        else:
            self.dwdt = matrix(0, (self.n, 1))
        if self.coi[0] is not None:
            self.copy_data_ext('COI', 'dwdt', 'dwdt_coi', self.coi)
        else:
            self.dwdt_coi = matrix(0, (self.n, 1))
        Turbine.servcall(self, dae)

    def init1(self, dae):
        self.servcall(dae)
        mva = self.system.mva
        self.p0 = mul(self.p0, 1)
        self.v120 = self.v12

        self.toMb = div(mva, self.Sn)  # to machine base
        self.toSb = self.Sn / mva  # to system base
        rs = matrix(self.rs)
        xd = matrix(self.xd)
        xq = matrix(self.xq)
        psip = matrix(self.psip)
        Pg = matrix(self.p0)

        # rotor speed
        omega = 1 * (ageb(mva * Pg, self.Sn)) + \
            mul(0.5 + 0.5 * mul(Pg, self.toMb),
                aandb(agtb(Pg, 0), altb(mva * Pg, self.Sn))) + \
            0.5 * (aleb(mva * Pg, 0))

        theta = mul(self.Kp, mround(1000 * (omega - 1)) / 1000)
        theta = mmax(theta, 0)

        # variables to initialize iteratively: vsd, vsq, isd, isq

        vsd = matrix(0.8, (self.n, 1))
        vsq = matrix(0.6, (self.n, 1))
        isd = matrix(self.p0 / 2)
        isq = matrix(self.p0 / 2)

        for i in range(self.n):
            # vsd = 0.5
            # vsq = self.psip[i]
            # isd = Pg / 2
            # isq = Pg / 2
            x = matrix([vsd[i], vsq[i], isd[i], isq[i]])

            mis = ones(4, 1)
            jac = sparse(matrix(0, (4, 4), 'd'))
            iter = 0
            while (max(abs(mis))) > self.system.tds.config.tol:
                if iter > 40:
                    logger.error(
                        'Initialization of WTG4DC <{}> failed.'.format(
                            self.name[i]))
                    break
                mis[0] = x[0] * x[2] + x[1] * x[3] - Pg[i]
                # mis[1] = omega[i]*x[3] * (psip[i] + (xq[i] - xd[i]) * x[2])\
                #     - Pg[i]
                mis[1] = omega[i] * x[3] * (psip[i] - xd[i] * x[2]) - Pg[i]
                mis[2] = -x[0] - rs[i] * x[2] + omega[i] * xq[i] * x[3]
                mis[3] = x[1] + rs[i] * x[3] + omega[i] * xd[i] * x[2] - \
                    omega[i] * psip[i]

                jac[0, 0] = x[2]
                jac[0, 1] = x[3]
                jac[0, 2] = x[0]
                jac[0, 3] = x[1]

                jac[1, 2] = omega[i] * x[3] * (-xd[i])
                jac[1, 3] = omega[i] * (psip[i] + (-xd[i]) * x[2])

                jac[2, 0] = -1
                jac[2, 2] = -rs[i]
                jac[2, 3] = omega[i] * xq[i]
                jac[3, 1] = 1
                jac[3, 2] = omega[i] * xd[i]
                jac[3, 3] = rs[i]

                linsolve(jac, mis)
                x -= mis
                iter += 1

            vsd[i] = x[0]
            vsq[i] = x[1]
            isd[i] = x[2]
            isq[i] = x[3]

        dae.y[self.isd] = isd
        dae.y[self.vsd] = vsd
        dae.y[self.vsq] = vsq

        dae.x[self.isq] = isq
        dae.x[self.omega_m] = mul(self.u0, omega)
        dae.x[self.theta_p] = mul(self.u0, theta)
        dae.y[self.pwa] = mmax(mmin(2 * dae.x[self.omega_m] - 1, 1), 0)

        self.ps0 = mul(vsd, isd) + mul(vsq, isq)
        self.qs0 = mul(vsq, isd) - mul(vsd, isq)
        self.te0 = mul(isq, psip + mul(xq - xd, isd))
        dae.y[self.te] = self.te0
        dae.y[self.ps] = self.ps0

        MPPT.init1(self, dae)
        Turbine.init1(self, dae)

        self.system.rmgen(self.dcgen)

    def gcall(self, dae):
        Turbine.gcall(self, dae)
        MPPT.gcall(self, dae)
        dae.g[self.isd] = -self.qs0 + mul(dae.y[self.isd],
                                          dae.y[self.vsq]) - mul(
                                              dae.x[self.isq], dae.y[self.vsd])
        dae.g[self.vsd] = -dae.y[self.vsd] - mul(
            dae.y[self.isd], self.rs) + mul(dae.x[self.isq],
                                            dae.x[self.omega_m], self.xq)
        dae.g[self.vsq] = -dae.y[self.vsq] - mul(
            dae.x[self.isq], self.rs) - mul(
                dae.x[self.omega_m],
                -self.psip + mul(dae.y[self.isd], self.xd))
        dae.g[self.ps] = -dae.y[self.ps] + mul(
            dae.y[self.isd], dae.y[self.vsd]) + mul(dae.x[self.isq],
                                                    dae.y[self.vsq])
        dae.g[self.te] = -dae.y[self.te] + mul(
            dae.x[self.isq],
            self.psip + mul(dae.y[self.isd], self.xq - self.xd))
        dae.g += spmatrix(
            -mul(dae.y[self.ps], div(1, dae.y[self.v1] - dae.y[self.v2])),
            self.v1, [0] * self.n, (dae.m, 1), 'd')
        dae.g += spmatrix(
            mul(dae.y[self.ps], div(1, dae.y[self.v1] - dae.y[self.v2])),
            self.v2, [0] * self.n, (dae.m, 1), 'd')

    def fcall(self, dae):
        Turbine.gcall(self, dae)
        dae.f[self.omega_m] = mul(
            0.5, div(1, self.H),
            -dae.y[self.te] + mul(dae.y[self.pw], div(1, dae.x[self.omega_m])))

        # dae.f[self.isq] = mul(
        #     div(1, self.Teq), -dae.x[self.isq] + mul(
        #         self.toSb, dae.y[self.pwa] - mul(self.Kdc, self.v12),
        #         div(1, dae.x[self.omega_m]),
        #         div(1, self.psip - mul(dae.y[self.isd], self.xd))))

        dae.f[self.isq] = mul(
            div(1, self.Teq), -dae.x[self.isq] + mul(
                self.toSb, div(1, dae.x[self.omega_m]),
                div(1, self.psip - mul(dae.y[self.isd], self.xd)),
                dae.y[self.pwa] - mul(self.Kcoi, dae.y[self.dwdt_coi]) - mul(
                    self.Kdc,
                    (dae.y[self.v1] - dae.y[self.v2]) - self.v120) - mul(
                        self.Ki, dae.y[self.dwdt])))

    @property
    def v12(self):
        dae = self.system.dae
        return dae.y[self.v1] - dae.y[self.v2]

    def gycall(self, dae):
        Turbine.gycall(self, dae)
        MPPT.gycall(self, dae)
        dae.add_jac(Gy, -dae.x[self.isq], self.isd, self.vsd)
        dae.add_jac(Gy, dae.y[self.vsq], self.isd, self.isd)
        dae.add_jac(Gy, dae.y[self.isd], self.isd, self.vsq)
        dae.add_jac(Gy, -mul(dae.x[self.omega_m], self.xd), self.vsq, self.isd)
        dae.add_jac(Gy, dae.y[self.isd], self.ps, self.vsd)
        dae.add_jac(Gy, dae.y[self.vsd], self.ps, self.isd)
        dae.add_jac(Gy, dae.x[self.isq], self.ps, self.vsq)
        dae.add_jac(Gy, mul(dae.x[self.isq], self.xq - self.xd), self.te,
                    self.isd)
        dae.add_jac(Gy, -div(1, dae.y[self.v1] - dae.y[self.v2]), self.v1,
                    self.ps)
        dae.add_jac(
            Gy, -mul(dae.y[self.ps], (dae.y[self.v1] - dae.y[self.v2])**-2),
            self.v1, self.v2)
        dae.add_jac(Gy,
                    mul(dae.y[self.ps], (dae.y[self.v1] - dae.y[self.v2])**-2),
                    self.v1, self.v1)
        dae.add_jac(Gy, div(1, dae.y[self.v1] - dae.y[self.v2]), self.v2,
                    self.ps)
        dae.add_jac(Gy,
                    mul(dae.y[self.ps], (dae.y[self.v1] - dae.y[self.v2])**-2),
                    self.v2, self.v2)
        dae.add_jac(
            Gy, -mul(dae.y[self.ps], (dae.y[self.v1] - dae.y[self.v2])**-2),
            self.v2, self.v1)

    def fxcall(self, dae):
        Turbine.jac0(self, dae)
        dae.add_jac(Gx, -dae.y[self.vsd], self.isd, self.isq)
        dae.add_jac(Gx, mul(dae.x[self.isq], self.xq), self.vsd, self.omega_m)
        dae.add_jac(Gx, mul(dae.x[self.omega_m], self.xq), self.vsd, self.isq)
        dae.add_jac(Gx, self.psip - mul(dae.y[self.isd], self.xd), self.vsq,
                    self.omega_m)
        dae.add_jac(Gx, dae.y[self.vsq], self.ps, self.isq)
        dae.add_jac(Gx, self.psip + mul(dae.y[self.isd], self.xq - self.xd),
                    self.te, self.isq)
        dae.add_jac(
            Fx,
            mul(-0.5, dae.y[self.pw], div(1, self.H), (dae.x[self.omega_m])
                ** -2), self.omega_m, self.omega_m)

        dae.add_jac(
            Fx, -mul(
                div(1, self.Teq), (dae.x[self.omega_m])**-2,
                div(1, self.psip - mul(dae.y[self.isd], self.xd)),
                dae.y[self.pwa] - mul(self.Kcoi, dae.y[self.dwdt_coi]) - mul(
                    self.Kdc, dae.y[self.v1] - dae.y[self.v2]) - mul(
                        self.Ki, dae.y[self.dwdt])), self.isq, self.omega_m)
        dae.add_jac(Fy, mul(0.5, div(1, self.H), div(1, dae.x[self.omega_m])),
                    self.omega_m, self.pw)

        dae.add_jac(
            Fy,
            mul(self.Kdc, div(1, self.Teq), div(1, dae.x[self.omega_m]),
                div(1, self.psip - mul(dae.y[self.isd], self.xd))), self.isq,
            self.v2)
        dae.add_jac(
            Fy, -mul(self.Ki, div(1, self.Teq), div(1, dae.x[self.omega_m]),
                     div(1, self.psip - mul(dae.y[self.isd], self.xd))),
            self.isq, self.dwdt)
        dae.add_jac(
            Fy,
            mul(
                div(1, self.Teq), div(1, dae.x[self.omega_m]),
                div(1, self.psip - mul(dae.y[self.isd], self.xd))), self.isq,
            self.pwa)
        dae.add_jac(
            Fy, -mul(self.Kdc, div(1, self.Teq), div(1, dae.x[self.omega_m]),
                     div(1, self.psip - mul(dae.y[self.isd], self.xd))),
            self.isq, self.v1)
        dae.add_jac(
            Fy,
            mul(
                self.xd, div(1, self.Teq), div(1, dae.x[self.omega_m]),
                (self.psip - mul(dae.y[self.isd], self.xd))**-2,
                dae.y[self.pwa] - mul(self.Kcoi, dae.y[self.dwdt_coi]) - mul(
                    self.Kdc, dae.y[self.v1] - dae.y[self.v2]) - mul(
                        self.Ki, dae.y[self.dwdt])), self.isq, self.isd)
        dae.add_jac(
            Fy, -mul(self.Kcoi, div(1, self.Teq), div(1, dae.x[self.omega_m]),
                     div(1, self.psip - mul(dae.y[self.isd], self.xd))),
            self.isq, self.dwdt_coi)

    def jac0(self, dae):
        Turbine.jac0(self, dae)
        MPPT.jac0(self, dae)
        dae.add_jac(Gy0, -1, self.vsd, self.vsd)
        dae.add_jac(Gy0, -self.rs, self.vsd, self.isd)
        dae.add_jac(Gy0, -1, self.vsq, self.vsq)
        dae.add_jac(Gy0, -1, self.ps, self.ps)
        dae.add_jac(Gy0, -1, self.te, self.te)
        dae.add_jac(Gx0, -self.rs, self.vsq, self.isq)
        dae.add_jac(Fx0, -div(1, self.Teq), self.isq, self.isq)
        dae.add_jac(Fy0, mul(-0.5, div(1, self.H)), self.omega_m, self.te)
        dae.add_jac(Gy0, 1e-6, self.isd, self.isd)
        dae.add_jac(Gy0, 1e-6, self.vsd, self.vsd)
        dae.add_jac(Gy0, 1e-6, self.vsq, self.vsq)
        dae.add_jac(Gy0, 1e-6, self.ps, self.ps)
        dae.add_jac(Gy0, 1e-6, self.te, self.te)


class WTG3(ModelBase):
    """Wind turbine type III"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'WTG'
        self._name = 'WTG3'
        self._algebs.extend([
            'isd', 'isq', 'vrd', 'vrq', 'vsd', 'vsq', 'vref', 'pwa', 'pw',
            'cp', 'lamb', 'ilamb'
        ])
        self._fnamex.extend(['\\theta_p', '\\omega_m', 'I_{r, d}', 'I_{r, q}'])
        self._fnamey.extend([
            'I_{s, d}', 'I_{s, q}', 'V_{r, d}', 'V_{r, q}', 'V_{s, d}',
            'V_{s, q}', 'V_{ref}', 'P_{\\omega a}', 'P_w', 'c_p', '\\lambda',
            '\\frac{1}{\\lambda}', '\\omega_{ref}'
        ])
        self._mandatory.extend(['bus', 'gen', 'wind'])
        self._params.extend([
            'fn', 'Kp', 'nblade', 'ngen', 'npole', 'R', 'Tp', 'Ts', 'ngb', 'H',
            'rr', 'rs', 'xr', 'xs', 'xmu', 'Te', 'KV', 'pmax', 'pmin', 'qmax',
            'qmin', 'gammap', 'gammaq'
        ])
        self._powers.extend(['H', 'pmax', 'pmin', 'qmax', 'qmin'])
        self._service.extend([
            'u0', 'vref0', 'irq_min', 'ird_min', 'phi', 'fn', 'ird_max', 'x0',
            'irq_max', 'pi', 'irq_off', 'mva_mega', 'x1'
        ])
        self._states.extend(['theta_p', 'omega_m', 'ird', 'irq'])
        self._times.extend(['Tp', 'Te'])
        self._z.extend(['rs', 'xs', 'rr', 'xr', 'xmu'])
        self._ac.update({'bus': ['a', 'v']})
        self._data.update({
            'fn': 60,
            'rs': 0.01,
            'xmu': 3,
            'R': 35,
            'ngb': 0.011235,
            'gammap': 1,
            'npole': 4,
            'qmin': -0.6,
            'KV': 10,
            'xr': 0.08,
            'Te': 0.01,
            'pmin': 0,
            'Ts': 1,
            'Sn': 40,
            'wind': 0,
            'gen': 0,
            'rr': 0.01,
            'pmax': 1.0,
            'gammaq': 1,
            'Kp': 10,
            'xs': 0.1,
            'H': 2,
            'Tp': 3.0,
            'qmax': 0.6,
            'nblade': 3,
            'bus': 0,
            'ngen': 40,
            'u': 1
        })
        self._descr.update({
            'fn': 'Base frequency',
            'rs': 'Stator resistance',
            'xmu': 'Magnetizing reactance',
            'R': 'Rotor radius',
            'pmax': 'Maximum active power',
            'gammap': 'Active power generation ratio',
            'npole': 'Number of poles',
            'qmin': 'Minimum reactive power',
            'KV': 'Voltage control gain',
            'xr': 'Rotor reactance',
            'Te': 'Power control time constant',
            'pmin': 'Minimum reactive power',
            'Ts': 'Speed control time constant',
            'wind': 'Wind time series idx',
            'gen': 'Static generator idx',
            'rr': 'Rotor resistance',
            'ngb': 'Gear box ratio',
            'gammaq': 'Reactive power generation ratio',
            'Kp': 'Pitch control gain',
            'xs': 'Stator reactance',
            'H': 'Machine rotor and turbine inertia constant',
            'Tp': 'Pitch control time constant',
            'qmax': 'Maximum active power',
            'nblade': 'Number of blades',
            'bus': 'Bus idx',
            'ngen': 'Number of generators'
        })
        self._units.update({
            'fn': 'Hz',
            'rs': 'pu',
            'xmu': 'pu',
            'rr': 'pu',
            'R': 'm',
            'pmax': 'pu',
            'qmin': 'pu',
            'Kp': 'pu',
            'xs': 'pu',
            'qmax': 'pu',
            'H': 'MWs/MVA',
            'Tp': 's',
            'KV': 'pu',
            'Te': 's',
            'xr': 'pu',
            'pmin': 'pu'
        })
        self.calls.update({
            'init1': True,
            'gycall': True,
            'fxcall': True,
            'fcall': True,
            'gcall': True,
            'jac0': True
        })
        self._init()

    def servcall(self, dae):
        self.copy_data_ext('StaticGen', 'u', 'ugen', self.gen)
        self.copy_data_ext('Bus', 'Pg', 'p0', self.bus)
        self.copy_data_ext('Bus', 'Qg', 'q0', self.bus)
        self.copy_data_ext('Wind', 'vw', 'vw', self.wind)
        self.copy_data_ext('Wind', 'rho', 'rho', self.wind)
        self.copy_data_ext('Wind', 'Vwn', 'Vwn', self.wind)
        self.vref0 = dae.y[self.v]
        self.x0 = self.xmu + self.xs
        self.x1 = self.xmu + self.xr
        self.ird_min = -999
        self.ird_max = 999
        self.irq_min = -999
        self.irq_max = 999
        self.irq_off = zeros(self.n, 1)
        self.u0 = mul(self.u, self.ugen)
        # self.omega_ref0 = ones(self.n, 1)
        self.mva_mega = 100000000.0

    def init1(self, dae):
        """New initialization function"""
        self.servcall(dae)
        retval = True

        mva = self.system.mva
        self.p0 = mul(self.p0, self.gammap)
        self.q0 = mul(self.q0, self.gammaq)

        dae.y[self.vsd] = mul(dae.y[self.v], -sin(dae.y[self.a]))
        dae.y[self.vsq] = mul(dae.y[self.v], cos(dae.y[self.a]))

        rs = matrix(self.rs)
        rr = matrix(self.rr)
        xmu = matrix(self.xmu)
        x1 = matrix(self.xs) + xmu
        x2 = matrix(self.xr) + xmu
        Pg = matrix(self.p0)
        Qg = matrix(self.q0)
        Vc = dae.y[self.v]
        vsq = dae.y[self.vsq]
        vsd = dae.y[self.vsd]

        toSn = div(mva, self.Sn)  # to machine base
        toSb = self.Sn / mva  # to system base

        # rotor speed
        omega = 1 * (ageb(mva * Pg, self.Sn)) + \
            mul(0.5 + 0.5 * mul(Pg, toSn),
                aandb(agtb(Pg, 0), altb(mva * Pg, self.Sn))) + \
            0.5 * (aleb(mva * Pg, 0))

        slip = 1 - omega
        theta = mul(self.Kp, mround(1000 * (omega - 1)) / 1000)
        theta = mmax(theta, 0)

        # prepare for the iterations

        irq = mul(-x1, toSb, (2 * omega - 1), div(1, Vc), div(1, xmu),
                  div(1, omega))
        isd = zeros(*irq.size)
        isq = zeros(*irq.size)

        # obtain ird isd isq
        for i in range(self.n):
            A = sparse([[-rs[i], vsq[i]], [x1[i], -vsd[i]]])
            B = matrix([vsd[i] - xmu[i] * irq[i], Qg[i]])
            linsolve(A, B)
            isd[i] = B[0]
            isq[i] = B[1]
        ird = -div(vsq + mul(rs, isq) + mul(x1, isd), xmu)
        vrd = -mul(rr, ird) + mul(
            slip,
            mul(x2, irq) + mul(xmu, isq))  # todo: check x1 or x2
        vrq = -mul(rr, irq) - mul(slip, mul(x2, ird) + mul(xmu, isd))

        # main iterations
        for i in range(self.n):
            mis = ones(6, 1)
            rows = [0, 0, 0, 1, 1, 1, 2, 2, 3, 3, 4, 4, 5]
            cols = [0, 1, 3, 0, 1, 2, 2, 4, 3, 5, 0, 1, 2]

            x = matrix([isd[i], isq[i], ird[i], irq[i], vrd[i], vrq[i]])
            # vals = [-rs, x1, xmu, -x1, -rs, -xmu, -rr,
            #         -1, -rr, -1, vsd, vsq, -xmu * Vc / x1]

            vals = [
                -rs[i], x1[i], xmu[i], -x1[i], -rs[i], -xmu[i], -rr[i], -1,
                -rr[i], -1, vsd[i], vsq[i], -xmu[i] * Vc[i] / x1[i]
            ]
            jac0 = spmatrix(vals, rows, cols, (6, 6), 'd')
            iter = 0

            while max(abs(mis)) > self.system.tds.config.tol:
                if iter > 20:
                    logger.error(
                        'Initialization of DFIG <{}> failed.'.format(
                            self.name[i]))
                    retval = False
                    break

                mis[0] = -rs[i] * x[0] + x1[i] * x[1] + xmu[i] * x[3] - vsd[i]
                mis[1] = -rs[i] * x[1] - x1[i] * x[0] - xmu[i] * x[2] - vsq[i]
                mis[2] = -rr[i] * x[2] + slip[i] * (
                    x2[i] * x[3] + xmu[i] * x[1]) - x[4]
                mis[3] = -rr[i] * x[3] - slip[i] * (
                    x2[i] * x[2] + xmu[i] * x[0]) - x[5]
                mis[4] = vsd[i] * x[0] + vsq[i] * x[1] + x[4] * x[2] + \
                    x[5] * x[3] - Pg[i]
                mis[5] = -xmu[i] * Vc[i] * x[2] / x1[i] - \
                    Vc[i] * Vc[i] / x1[i] - Qg[i]

                rows = [2, 2, 3, 3, 4, 4, 4, 4]
                cols = [1, 3, 0, 2, 2, 3, 4, 5]
                vals = [
                    slip[i] * xmu[i], slip[i] * x2[i], -slip[i] * xmu[i],
                    -slip[i] * x2[i], x[4], x[5], x[2], x[3]
                ]

                jac = jac0 + spmatrix(vals, rows, cols, (6, 6), 'd')

                linsolve(jac, mis)

                x -= mis
                iter += 1

            isd[i] = x[0]
            isq[i] = x[1]
            ird[i] = x[2]
            irq[i] = x[3]
            vrd[i] = x[4]
            vrq[i] = x[5]

        dae.x[self.ird] = mul(self.u0, ird)
        dae.x[self.irq] = mul(self.u0, irq)
        dae.y[self.isd] = isd
        dae.y[self.isq] = isq
        dae.y[self.vrd] = vrd
        dae.y[self.vrq] = vrq

        dae.x[self.omega_m] = mul(self.u0, omega)
        dae.x[self.theta_p] = mul(self.u0, theta)
        dae.y[self.pwa] = mmax(mmin(2 * dae.x[self.omega_m] - 1, 1), 0)

        self.vref0 = mul(
            aneb(self.KV, 0), Vc - div(ird + div(Vc, xmu), self.KV))
        dae.y[self.vref] = self.vref0
        k = mul(div(x1, Vc, xmu, omega), toSb)

        self.irq_off = -mul(k, mmax(mmin(2 * omega - 1, 1), 0)) - irq

        # electrical torque in pu
        te = mul(
            xmu,
            mul(dae.x[self.irq], dae.y[self.isd]) - mul(
                dae.x[self.ird], dae.y[self.isq]))

        for i in range(self.n):
            if te[i] < 0:
                logger.error(
                    'Pe < 0 on bus <{}>. Wind speed initialize failed.'.
                    format(self.bus[i]))
                retval = False

        # wind power in pu
        pw = mul(te, omega)
        dae.y[self.pw] = pw

        # wind speed initialization loop

        R = 4 * pi * self.system.freq * mul(self.R, self.ngb,
                                            div(1, self.npole))
        AA = pi * self.R**2
        vw = 0.9 * self.Vwn

        for i in range(self.n):
            mis = 1
            iter = 0
            while abs(mis) > self.system.tds.config.tol:
                if iter > 50:
                    logger.error(
                        'Wind <{}> init failed. '
                        'Try increasing the nominal wind speed.'.
                        format(self.wind[i]))
                    retval = False
                    break

                pw_iter, jac = self.windpower(self.ngen[i], self.rho[i], vw[i],
                                              AA[i], R[i], omega[i], theta[i])

                mis = pw_iter - pw[i]
                inc = -mis / jac[1]
                vw[i] += inc
                iter += 1

        # set wind speed
        dae.x[self.vw] = div(vw, self.Vwn)

        lamb = div(omega, vw, div(1, R))
        ilamb = div(1,
                    (div(1, lamb + 0.08 * theta) - div(0.035, theta**3 + 1)))
        cp = 0.22 * mul(
            div(116, ilamb) - 0.4 * theta - 5, exp(div(-12.5, ilamb)))

        dae.y[self.lamb] = lamb
        dae.y[self.ilamb] = ilamb
        dae.y[self.cp] = cp

        self.system.rmgen(self.gen)

        if not retval:
            logger.error('DFIG initialization failed')

        return retval

    def windpower(self, ngen, rho, vw, Ar, R, omega, theta, derivative=False):
        mva_mega = self.system.mva * 1e6
        lamb = omega * R / vw
        ilamb = 1 / (1 / (lamb + 0.08 * theta) - 0.035 / (theta**3 + 1))
        cp = 0.22 * (116 / ilamb - 0.4 * theta - 5) * exp(-12.5 / ilamb)
        pw = 0.5 * ngen * rho * cp * Ar * vw**3 / mva_mega

        a1 = exp(-12.5 / ilamb)
        a2 = (lamb + 0.08 * theta)**2
        a3 = 116. / ilamb - 0.4 * theta - 5
        a4 = -9.28 / (lamb + 0.08 * theta) ** 2 + \
            12.180 * theta * theta / (theta ** 3 + 1) ** 2 - 0.4
        a5 = 1.000 / (lamb + 0.08 * theta) ** 2 - \
            1.3125 * theta * theta / (theta ** 3 + 1) ** 2

        jac = ones(1, 3)
        jac[0] = ngen * R * a1 * rho * vw * vw * Ar * (
            -12.760 + 1.3750 * a3) / a2 / mva_mega
        jac[1] = ngen * (omega * R * (12.760 - 1.3750 * a3) / a2 +
                         0.330 * a3 * vw) * vw * Ar * rho * a1 / mva_mega
        jac[2] = ngen * 0.110 * rho * (
            a4 + a3 * a5) * a1 * Ar * vw**3 / mva_mega

        return pw, jac

    @property
    def phi(self):
        deg1 = pi / 180
        dae = self.system.dae
        above = agtb(dae.x[self.omega_m], 1)
        phi_degree_step = mfloor((dae.x[self.omega_m] - 1) / deg1) * deg1
        return mul(phi_degree_step, above)

    def gcall(self, dae):
        dae.g[self.isd] = -dae.y[self.vsd] + mul(
            dae.x[self.irq], self.xmu) + mul(dae.y[self.isq], self.x0) - mul(
                dae.y[self.isd], self.rs)
        dae.g[self.isq] = -dae.y[self.vsq] - mul(
            dae.x[self.ird], self.xmu) - mul(dae.y[self.isd], self.x0) - mul(
                dae.y[self.isq], self.rs)
        dae.g[self.vrd] = -dae.y[self.vrd] + mul(
            1 - dae.x[self.omega_m],
            mul(dae.x[self.irq], self.x1) + mul(
                dae.y[self.isq], self.xmu)) - mul(dae.x[self.ird], self.rr)
        dae.g[self.vrq] = -dae.y[self.vrq] - mul(
            dae.x[self.irq], self.rr) - mul(
                1 - dae.x[self.omega_m],
                mul(dae.x[self.ird], self.x1) + mul(dae.y[self.isd], self.xmu))
        dae.g[self.vsd] = -dae.y[self.vsd] - mul(dae.y[self.v],
                                                 sin(dae.y[self.a]))
        dae.g[self.vsq] = -dae.y[self.vsq] + mul(dae.y[self.v],
                                                 cos(dae.y[self.a]))
        dae.g[self.vref] = self.vref0 - dae.y[self.vref]
        dae.g[self.pwa] = mmax(mmin(2 * dae.x[self.omega_m] - 1, 1),
                               0) - dae.y[self.pwa]

        dae.hard_limit(self.pwa, 0, 1)
        dae.g[self.pw] = -dae.y[self.pw] + mul(
            0.5, dae.y[self.cp], self.ngen, pi, self.rho, (self.R)**2,
            (self.Vwn)**3, div(1, self.mva_mega), (dae.x[self.vw])**3)
        dae.g[self.cp] = -dae.y[self.cp] + mul(
            -1.1 + mul(25.52, div(1, dae.y[self.ilamb])) + mul(
                -0.08800000000000001, dae.x[self.theta_p]),
            exp(mul(-12.5, div(1, dae.y[self.ilamb]))))
        dae.g[self.lamb] = -dae.y[self.lamb] + mul(
            4, self.R, self.fn, self.ngb, dae.x[self.omega_m], pi,
            div(1, self.Vwn), div(1, self.npole), div(1, dae.x[self.vw]))
        dae.g[self.ilamb] = div(
            1,
            div(1, dae.y[self.lamb] + mul(0.08, dae.x[self.theta_p])) + mul(
                -0.035, div(1, 1 +
                            (dae.x[self.theta_p])**3))) - dae.y[self.ilamb]
        dae.g += spmatrix(
            mul(
                self.u0, -mul(dae.x[self.ird], dae.y[self.vrd]) - mul(
                    dae.x[self.irq], dae.y[self.vrq]) - mul(
                        dae.y[self.isd], dae.y[self.vsd]) - mul(
                            dae.y[self.isq], dae.y[self.vsq])), self.a,
            [0] * self.n, (dae.m, 1), 'd')
        dae.g += spmatrix(
            mul(
                self.u0,
                mul((dae.y[self.v])**2, div(1, self.x0)) + mul(
                    dae.x[self.ird], dae.y[self.v], self.xmu, div(
                        1, self.x0))), self.v, [0] * self.n, (dae.m, 1), 'd')

    def fcall(self, dae):
        toSb = self.Sn / self.system.mva
        omega = not0(dae.x[self.omega_m])
        dae.f[self.theta_p] = mul(
            div(1, self.Tp), -dae.x[self.theta_p] + mul(
                self.Kp, self.phi, -1 + dae.x[self.omega_m]))
        dae.anti_windup(self.theta_p, 0, pi)

        dae.f[self.omega_m] = mul(
            0.5, div(1, self.H),
            mul(dae.y[self.pw], div(1, dae.x[self.omega_m])) - mul(
                self.xmu,
                mul(dae.x[self.irq], dae.y[self.isd]) - mul(
                    dae.x[self.ird], dae.y[self.isq])))

        dae.f[self.ird] = mul(
            div(1, self.Ts),
            -dae.x[self.ird] + mul(self.KV, dae.y[self.v] - dae.y[self.vref]) -
            mul(dae.y[self.v], div(1, self.xmu)))
        dae.anti_windup(self.ird, self.ird_min, self.irq_max)
        k = mul(self.x0, toSb, div(1, dae.y[self.v]), div(1, self.xmu),
                div(1, omega))
        dae.f[self.irq] = mul(
            div(1, self.Te),
            -dae.x[self.irq] - self.irq_off - mul(dae.y[self.pwa], k))
        dae.anti_windup(self.irq, self.irq_min, self.irq_max)

    def gycall(self, dae):
        dae.add_jac(Gy, mul(self.xmu, 1 - dae.x[self.omega_m]), self.vrd,
                    self.isq)
        dae.add_jac(Gy, -mul(self.xmu, 1 - dae.x[self.omega_m]), self.vrq,
                    self.isd)
        dae.add_jac(Gy, -sin(dae.y[self.a]), self.vsd, self.v)
        dae.add_jac(Gy, -mul(dae.y[self.v], cos(dae.y[self.a])), self.vsd,
                    self.a)
        dae.add_jac(Gy, cos(dae.y[self.a]), self.vsq, self.v)
        dae.add_jac(Gy, -mul(dae.y[self.v], sin(dae.y[self.a])), self.vsq,
                    self.a)
        dae.add_jac(
            Gy,
            mul(0.5, self.ngen, pi, self.rho, (self.R)**2, (self.Vwn)**3,
                div(1, self.mva_mega), (dae.x[self.vw])**3), self.pw, self.cp)
        dae.add_jac(
            Gy,
            mul(-25.52, (dae.y[self.ilamb])**-2,
                exp(mul(-12.5, div(1, dae.y[self.ilamb])))) + mul(
                    12.5, (dae.y[self.ilamb])**-2,
                    -1.1 + mul(25.52, div(1, dae.y[self.ilamb])) + mul(
                        -0.088, dae.x[self.theta_p]),
                    exp(mul(-12.5, div(1, dae.y[self.ilamb])))), self.cp,
            self.ilamb)
        dae.add_jac(
            Gy,
            mul((dae.y[self.lamb] + mul(0.08, dae.x[self.theta_p]))**-2,
                (div(1, dae.y[self.lamb] + mul(0.08, dae.x[self.theta_p])) +
                 mul(-0.035, div(1, 1 + (dae.x[self.theta_p])**3)))**-2),
            self.ilamb, self.lamb)
        dae.add_jac(Gy, -mul(dae.y[self.isd], self.u0), self.a, self.vsd)
        dae.add_jac(Gy, -mul(dae.x[self.irq], self.u0), self.a, self.vrq)
        dae.add_jac(Gy, -mul(self.u0, dae.y[self.vsq]), self.a, self.isq)
        dae.add_jac(Gy, -mul(dae.x[self.ird], self.u0), self.a, self.vrd)
        dae.add_jac(Gy, -mul(dae.y[self.isq], self.u0), self.a, self.vsq)
        dae.add_jac(Gy, -mul(self.u0, dae.y[self.vsd]), self.a, self.isd)
        dae.add_jac(
            Gy,
            mul(
                self.u0,
                mul(2, dae.y[self.v], div(1, self.x0)) + mul(
                    dae.x[self.ird], self.xmu, div(1, self.x0))), self.v,
            self.v)

    def fxcall(self, dae):
        omega = not0(dae.x[self.omega_m])
        toSb = div(self.Sn, self.system.mva)
        dae.add_jac(Gx, mul(self.x1, 1 - dae.x[self.omega_m]), self.vrd,
                    self.irq)
        dae.add_jac(
            Gx,
            -mul(dae.x[self.irq], self.x1) - mul(dae.y[self.isq], self.xmu),
            self.vrd, self.omega_m)
        dae.add_jac(
            Gx,
            mul(dae.x[self.ird], self.x1) + mul(dae.y[self.isd], self.xmu),
            self.vrq, self.omega_m)
        dae.add_jac(Gx, -mul(self.x1, 1 - dae.x[self.omega_m]), self.vrq,
                    self.ird)
        dae.add_jac(
            Gx,
            mul(1.5, dae.y[self.cp],
                self.ngen, pi, self.rho, (self.R)**2, (self.Vwn)**3,
                div(1, self.mva_mega), (dae.x[self.vw])**2), self.pw, self.vw)
        dae.add_jac(
            Gx,
            mul(-0.088, exp(
                mul(-12.5, div(1, dae.y[self.ilamb])))), self.cp, self.theta_p)
        dae.add_jac(
            Gx,
            mul(-4, self.R, self.fn, self.ngb, dae.x[self.omega_m], pi,
                div(1, self.Vwn), div(1, self.npole), (dae.x[self.vw])**-2),
            self.lamb, self.vw)
        dae.add_jac(
            Gx,
            mul(4, self.R, self.fn, self.ngb, pi, div(1, self.Vwn),
                div(1, self.npole), div(1, dae.x[self.vw])), self.lamb,
            self.omega_m)
        dae.add_jac(
            Gx,
            mul((div(1, dae.y[self.lamb] + mul(0.08, dae.x[self.theta_p])) +
                 mul(-0.035, div(1, 1 + (dae.x[self.theta_p]) ** 3))) ** -2,
                mul(0.08, (dae.y[self.lamb] + mul(0.08, dae.x[self.theta_p]))
                    ** -2) + mul(-0.105, (dae.x[self.theta_p]) ** 2,
                                 (1 + (dae.x[self.theta_p]) ** 3) ** -2)),
            self.ilamb,
            self.theta_p)
        dae.add_jac(Gx, -mul(self.u0, dae.y[self.vrq]), self.a, self.irq)
        dae.add_jac(Gx, -mul(self.u0, dae.y[self.vrd]), self.a, self.ird)
        dae.add_jac(Gx, mul(self.u0, dae.y[self.v], self.xmu, div(1, self.x0)),
                    self.v, self.ird)

        dae.add_jac(
            Fx,
            mul(dae.y[self.pwa], self.x0, toSb, div(1, self.Te),
                (dae.x[self.omega_m])**-2, div(1, dae.y[self.v]),
                div(1, self.xmu)), self.irq, self.omega_m)
        dae.add_jac(
            Fy,
            mul(dae.y[self.pwa], self.x0, toSb, div(1, self.Te), div(1, omega),
                (dae.y[self.v])**-2, div(1, self.xmu)), self.irq, self.v)
        dae.add_jac(
            Fy, -mul(self.x0, toSb, div(1, self.Te), div(1, omega),
                     div(1, dae.y[self.v]), div(1, self.xmu)), self.irq,
            self.pwa)

        dae.add_jac(Fx, mul(0.5, dae.y[self.isq], self.xmu, div(1, self.H)),
                    self.omega_m, self.ird)
        dae.add_jac(
            Fx,
            mul(-0.5, dae.y[self.pw], div(1, self.H), (dae.x[self.omega_m])
                ** -2), self.omega_m, self.omega_m)
        dae.add_jac(Fx, mul(-0.5, dae.y[self.isd], self.xmu, div(1, self.H)),
                    self.omega_m, self.irq)
        dae.add_jac(Fy, mul(0.5, div(1, self.H), div(1, dae.x[self.omega_m])),
                    self.omega_m, self.pw)
        dae.add_jac(Fy, mul(-0.5, dae.x[self.irq], self.xmu, div(1, self.H)),
                    self.omega_m, self.isd)
        dae.add_jac(Fy, mul(0.5, dae.x[self.ird], self.xmu, div(1, self.H)),
                    self.omega_m, self.isq)

    def jac0(self, dae):
        dae.add_jac(Gy0, -1, self.isd, self.vsd)
        dae.add_jac(Gy0, -self.rs, self.isd, self.isd)
        dae.add_jac(Gy0, self.x0, self.isd, self.isq)
        dae.add_jac(Gy0, -1, self.isq, self.vsq)
        dae.add_jac(Gy0, -self.x0, self.isq, self.isd)
        dae.add_jac(Gy0, -self.rs, self.isq, self.isq)
        dae.add_jac(Gy0, -1, self.vrd, self.vrd)
        dae.add_jac(Gy0, -1, self.vrq, self.vrq)
        dae.add_jac(Gy0, -1, self.vsd, self.vsd)
        dae.add_jac(Gy0, -1, self.vsq, self.vsq)
        dae.add_jac(Gy0, -1, self.vref, self.vref)
        dae.add_jac(Gy0, -1, self.pwa, self.pwa)
        dae.add_jac(Gy0, -1, self.pw, self.pw)
        dae.add_jac(Gy0, -1, self.cp, self.cp)
        dae.add_jac(Gy0, -1, self.lamb, self.lamb)
        dae.add_jac(Gy0, -1, self.ilamb, self.ilamb)
        dae.add_jac(Gx0, self.xmu, self.isd, self.irq)
        dae.add_jac(Gx0, -self.xmu, self.isq, self.ird)
        dae.add_jac(Gx0, -self.rr, self.vrd, self.ird)
        dae.add_jac(Gx0, -self.rr, self.vrq, self.irq)
        dae.add_jac(Gx0, 2, self.pwa, self.omega_m)
        dae.add_jac(Fx0, -div(1, self.Tp), self.theta_p, self.theta_p)
        dae.add_jac(Fx0, mul(self.Kp, self.phi, div(1, self.Tp)), self.theta_p,
                    self.omega_m)
        dae.add_jac(Fx0, -div(1, self.Ts), self.ird, self.ird)
        dae.add_jac(Fx0, -div(1, self.Te), self.irq, self.irq)
        dae.add_jac(Fy0, -mul(self.KV, div(1, self.Ts)), self.ird, self.vref)
        dae.add_jac(Fy0, mul(div(1, self.Ts), self.KV - div(1, self.xmu)),
                    self.ird, self.v)
        dae.add_jac(Gy0, 1e-6, self.isd, self.isd)
        dae.add_jac(Gy0, 1e-6, self.isq, self.isq)
        dae.add_jac(Gy0, 1e-6, self.vrd, self.vrd)
        dae.add_jac(Gy0, 1e-6, self.vrq, self.vrq)
        dae.add_jac(Gy0, 1e-6, self.vsd, self.vsd)
        dae.add_jac(Gy0, 1e-6, self.vsq, self.vsq)
        dae.add_jac(Gy0, 1e-6, self.vref, self.vref)
        dae.add_jac(Gy0, 1e-6, self.pwa, self.pwa)
        dae.add_jac(Gy0, 1e-6, self.pw, self.pw)
        dae.add_jac(Gy0, 1e-6, self.cp, self.cp)
        dae.add_jac(Gy0, 1e-6, self.lamb, self.lamb)
        dae.add_jac(Gy0, 1e-6, self.ilamb, self.ilamb)



import logging

logger = logging.getLogger(__name__)


class VarName(object):
    """Variable name manager class"""

    def __init__(self, system):
        self.system = system
        self.unamex = []  # unformatted state variable names
        self.unamey = []  # unformatted algeb variable names
        self.fnamex = []  # formatted state variable names
        self.fnamey = []  # formatted algeb variable names

    def resize(self):
        """Resize (extend) the list for variable names"""
        yext = self.system.dae.m - len(self.unamey)
        xext = self.system.dae.n - len(self.unamex)
        if yext > 0:
            self.unamey.extend([''] * yext)
            self.fnamey.extend([''] * yext)
        if xext > 0:
            self.unamex.extend([''] * xext)
            self.fnamex.extend([''] * xext)

    def resize_for_flows(self):
        """Extend `unamey` and `fnamey` for bus injections and line flows"""
        if self.system.config.dime_enable:
            self.system.tds.config.compute_flows = True

        if self.system.tds.config.compute_flows:
            nflows = 2 * self.system.Bus.n + \
                     8 * self.system.Line.n + \
                     2 * self.system.Area.n_combination
            self.unamey.extend([''] * nflows)
            self.fnamey.extend([''] * nflows)

    def append(self, listname, xy_idx, var_name, element_name):
        """Append variable names to the name lists"""
        self.resize()
        string = '{0} {1}'
        if listname not in ['unamex', 'unamey', 'fnamex', 'fnamey']:
            logger.error('Wrong list name for varname.')
            return
        elif listname in ['fnamex', 'fnamey']:
            string = '${0}\\ {1}$'

        if isinstance(element_name, list):
            for i, j in zip(xy_idx, element_name):
                # manual elem_add LaTex space for auto-generated element name
                if listname == 'fnamex' or listname == 'fnamey':
                    j = j.replace(' ', '\\ ')
                self.__dict__[listname][i] = string.format(var_name, j)
        elif isinstance(element_name, int):
            self.__dict__[listname][xy_idx] = string.format(
                var_name, element_name)
        else:
            logger.warning(
                'Unknown element_name type while building varname')

    def bus_line_names(self):
        """Append bus injection and line flow names to `varname`"""
        if self.system.tds.config.compute_flows:
            self.system.Bus._varname_inj()
            self.system.Line._varname_flow()
            self.system.Area._varname_inter()

    @property
    def uname(self):
        """
        Return the full unformatted variable name list following the order of
        state vars, algeb vars and line flow vars

        :return: list of unformatted names
        """
        return self.unamex + self.unamey

    @property
    def fname(self):
        """
        Return the full formatted variable name list following the order of
        state vars, algeb vars and line flow vars

        :return: list of formatted names
        """
        return self.fnamex + self.fnamey

    def get_xy_name(self, yidx, xidx=0):
        """
        Return variable names for the given indices

        :param yidx:
        :param xidx:
        :return:
        """
        assert isinstance(xidx, int)
        if isinstance(yidx, int):
            yidx = [yidx]

        uname = ['Time [s]'] + self.uname
        fname = ['$Time\\ [s]$'] + self.fname

        xname = [list(), list()]
        yname = [list(), list()]

        xname[0] = uname[xidx]
        xname[1] = fname[xidx]

        yname[0] = [uname[i] for i in yidx]
        yname[1] = [fname[i] for i in yidx]

        return xname, yname



from . import ConfigBase
from ..utils.cached import cached
import logging

logger = logging.getLogger(__name__)
try:
    from cvxoptklu import klu  # NOQA
    KLU = True
except ImportError:
    KLU = False


class System(ConfigBase):
    def __init__(self, **kwargs):
        self.freq = 60.0
        self.mva = 100.0
        self.distrsw = False
        self.sparselib = 'klu'
        self.sparselib_alt = ['klu', 'umfpack']
        self.export = 'txt'
        self.export_alt = ['txt', 'latex']
        self.coi = False
        self.connectivity = False
        self.error = 1
        self.static = 0
        self.nseries = 0
        self.forcepq = False
        self.forcez = False
        self.base = True
        self.dime_enable = False
        self.dime_name = 'sim'
        self.dime_server = 'ipc:///tmp/dime'
        super(System, self).__init__(**kwargs)

    @cached
    def config_descr(self):
        descriptions = {
            'freq': 'system base frequency',
            'mva': 'system base MVA',
            'distrsw': 'use distributed slack bus mode',
            'sparselib': 'sparse matrix library name',
            'export': 'help documentation export format',
            'coi': 'using Center of Inertia',
            'connectivity': 'connectivity check during TDS',
            'forcepq': 'force to use constant PQ load',
            'forcez': 'force to convert load to impedance',
            'base': 'convert  parameters to the system base',
        }
        return descriptions

    def check(self):
        """
        Check config data consistency

        Returns
        -------

        """
        if self.sparselib not in self.sparselib_alt:
            logger.warning("Invalid sparse library <{}>".format(self.sparselib))
            self.sparselib = 'umfpack'

        if self.sparselib == 'klu' and not KLU:
            logger.info("Optional package \"cvxoptklu\" available for speed up")
            self.sparselib = 'umfpack'

        return True



import numpy as np

from cvxopt import matrix
import logging

logger = logging.getLogger(__name__)


class VarOut(object):
    """
    Output variable value recorder

    TODO: merge in tds.py
    """

    def __init__(self, system):
        """Constructor of empty Varout object"""
        self.system = system
        self.t = []
        self.k = []
        self.vars = []
        self.vars_array = None
        self.dat = None
        self._mode = 'w'

    def store(self, t, step):
        """
        Record the state/algeb values at time t to self.vars
        """
        max_cache = int(self.system.tds.config.max_cache)
        if len(self.vars) >= max_cache > 0:
            self.dump()
            self.vars = list()
            self.t = list()
            self.k = list()
            logger.debug(
                'varout cache cleared at simulation t = {:g}.'.format(
                    self.system.dae.t))
            self._mode = 'a'

        self.t.append(t)
        self.k.append(step)
        self.vars.append(matrix([self.system.dae.x, self.system.dae.y]))

        # remove the post-computed variables from the variable list
        if self.system.tds.config.compute_flows:
            self.system.dae.y = self.system.dae.y[:self.system.dae.m]

    def show(self):
        """
        The representation of an Varout object

        :return: the full result matrix (for use with PyCharm viewer)
        :rtype: np.array
        """
        out = []

        for item in self.vars:
            out.append(list(item))

        return np.array(out)

    def concat_t_vars(self):
        """
        Concatenate ``self.t`` with ``self.vars`` and output a single matrix
        for data dump

        :return matrix: concatenated matrix with ``self.t`` as the 0-th column
        """
        out = np.array([])

        if len(self.t) == 0:
            return out

        out = np.ndarray(shape=(0, self.vars[0].size[0] + 1))

        for t, var in zip(self.t, self.vars):
            line = [[t]]
            line[0].extend(list(var))
            out = np.append(out, line, axis=0)

        return out

    def get_xy(self, yidx, xidx=0):
        """
        Return stored data for the given indices for plot

        :param yidx: the indices of the y-axis variables(1-indexing)
        :param xidx: the index of the x-axis variables
        :return: None
        """
        assert isinstance(xidx, int)
        if isinstance(yidx, int):
            yidx = [yidx]

        t_vars = self.concat_t_vars()

        xdata = t_vars[:, xidx]
        ydata = t_vars[:, yidx]

        return xdata.tolist(), ydata.transpose().tolist()

    def dump(self):
        """
        Dump the TDS results to the output `dat` file

        :param lst: dump ``.lst`` file

        :return: succeed flag
        """
        ret = False

        if self.system.files.no_output:
            # return ``True`` because it did not fail
            return True

        if self.write_lst() and self.write_dat():
            ret = True

        return ret

    def write_dat(self):
        """
        Write ``system.Varout.vars`` to a ``.dat`` file
        :return:
        """
        ret = False
        system = self.system

        # compute the total number of columns, excluding time
        if not system.Recorder.n:
            n_vars = system.dae.m + system.dae.n
            # post-computed power flows include:
            #   bus   - (Pi, Qi)
            #   line  - (Pij, Pji, Qij, Qji, Iij_Real, Iij_Imag, Iji_real, Iji_Imag)
            if system.tds.config.compute_flows:
                n_vars += 2 * system.Bus.n + 8 * system.Line.n + 2 * system.Area.n_combination
            idx = list(range(n_vars))

        else:
            n_vars = len(system.Recorder.varout_idx)
            idx = system.Recorder.varout_idx

        template = ['{:<8g}'] + ['{:0.10f}'] * n_vars
        template = ' '.join(template)

        # format the output in a string
        out = ''
        for t, line in zip(self.t, self.vars):
            values = [t] + list(line[idx])
            out += template.format(*values) + '\n'

        try:
            with open(system.files.dat, self._mode) as f:
                f.write(out)
            ret = True

        except IOError:
            logger.error('I/O Error while writing the dat file.')

        return ret

    def write_lst(self):
        """
        Dump the variable name lst file

        :return: succeed flag
        """

        ret = False
        out = ''
        system = self.system
        dae = self.system.dae
        varname = self.system.varname
        template = '{:>6g}, {:>25s}, {:>35s}\n'

        # header line
        out += template.format(0, 'Time [s]', '$Time\\ [s]$')

        # include line flow variables in algebraic variables
        nflows = 0
        if self.system.tds.config.compute_flows:
            nflows = 2 * self.system.Bus.n + \
                     8 * self.system.Line.n + \
                     2 * self.system.Area.n_combination

        # output variable indices
        if system.Recorder.n == 0:
            state_idx = list(range(dae.n))
            algeb_idx = list(range(dae.n, dae.n + dae.m + nflows))
            idx = state_idx + algeb_idx
        else:
            idx = system.Recorder.varout_idx

        # variable names concatenated
        uname = varname.unamex + varname.unamey
        fname = varname.fnamex + varname.fnamey

        for e, i in enumerate(idx):
            out += template.format(e + 1, uname[i], fname[i])

        try:
            with open(self.system.files.lst, 'w') as f:
                f.write(out)
            ret = True
        except IOError:
            logger.error('I/O Error while writing the lst file.')

        return ret

    def vars_to_array(self):
        """
        Convert `self.vars` to a numpy array

        Returns
        -------
        numpy.array
        """
        if not self.vars:
            return None

        vars_matrix = matrix(self.vars, size=(self.vars[0].size[0],
                                              len(self.vars))).trans()

        self.vars_array = np.array(vars_matrix)

        return self.vars_array



import sys

import pprint
try:
    from sympy import Symbol, diff, sin, cos, exp, Integer  # NOQA
except ImportError:
    raise ImportError('Please install sympy to parse ANDES cards.')

from andes.utils.math import to_number
import logging

logger = logging.getLogger(__name__)


def testlines(fid):
    try:
        first = fid.readline()
        if 'card' in first:
            return True
        else:
            return False
    except IOError:
        print('* IOError while reading input card file.')


def read(file, system):
    """Parse an ANDES card file into internal variables"""
    try:
        fid = open(file, 'r')
        raw_file = fid.readlines()
    except IOError:
        print('* IOError while reading input card file.')
        return

    ret_dict = dict()
    ret_dict['outfile'] = file.split('.')[0].lower() + '.py'
    key, val = None, None
    for idx, line in enumerate(raw_file):
        line = line.strip()
        if not line:
            continue
        if line.startswith('#'):
            continue
        elif '#' in line:
            line = line.split('#')[0]
        if '=' in line:  # defining a field
            key, val = line.split('=')
            key, val = key.strip(), val.strip()
            val = [] if val == '' else val
            ret_dict.update({key: val})
            if val:
                val = val.split(';')
        else:
            val.extend(line.split(';'))
        if val:
            val = de_blank(val)
            ret_dict[key] = val

    ret_dict_ord = dict(ret_dict)

    for key, val in ret_dict.items():
        if not val:
            continue
        if type(val) == list:
            if ':' in val[0]:
                new_val = {}  # return in a dictionary
                new_val_ord = [
                ]  # return in an ordered list with the dict keys at 0
                for item in val:
                    try:
                        m, n = item.split(':')
                    except ValueError:
                        print('* Error: check line <{}>'.format(item))
                        return
                    m, n = m.strip(), n.strip()
                    if ',' in n:
                        n = n.split(',')
                        n = de_blank(n)
                        n = [to_number(i) for i in n]
                    else:
                        n = to_number(n)
                    new_val.update({m.strip(): n})
                    new_val_ord.append([m.strip(), n])
                ret_dict[key] = new_val
                ret_dict_ord[key] = new_val_ord

    ret_dict['name'] = ret_dict['name'][0]
    ret_dict['doc_string'] = ret_dict['doc_string'][0]
    ret_dict['group'] = ret_dict['group'][0]
    ret_dict['service_keys'] = list(ret_dict['service_eq'].keys())
    ret_dict['consts'] = list(ret_dict['data'].keys()) + list(
        ret_dict['service_eq'].keys())
    ret_dict['init1_eq'] = ret_dict_ord['init1_eq']
    ret_dict['service_eq'] = ret_dict_ord['service_eq']
    ret_dict['ctrl'] = ret_dict_ord['ctrl']

    copy_algebs = []
    copy_states = []
    for item in ret_dict['ctrl']:
        key, val = item
        if val[3] == 'y':
            copy_algebs.append(key)
        elif val[3] == 'x':
            copy_states.append(key)
        elif val[3] == 'c':
            ret_dict['consts'].append(key)
    ret_dict['copy_algebs'] = copy_algebs
    ret_dict['copy_states'] = copy_states

    return run(system, **ret_dict)


def add_quotes(string):
    return '\'{}\''.format(string)


def de_blank(val):
    """Remove blank elements in `val` and return `ret`"""
    ret = list(val)
    if type(val) == list:
        for idx, item in enumerate(val):
            if item.strip() == '':
                ret.remove(item)
            else:
                ret[idx] = item.strip()
    return ret


def run(system,
        outfile='',
        name='',
        doc_string='',
        group='',
        data={},
        descr={},
        units={},
        params=[],
        fnamex=[],
        fnamey=[],
        mandatory=[],
        zeros=[],
        powers=[],
        currents=[],
        voltages=[],
        z=[],
        y=[],
        dccurrents=[],
        dcvoltages=[],
        r=[],
        g=[],
        times=[],
        ac={},
        dc={},
        ctrl={},
        consts=[],
        algebs=[],
        interfaces=[],
        states=[],
        init1_eq={},
        service_eq={},
        algeb_eq=[],
        windup={},
        hard_limit={},
        diff_eq=[],
        anti_windup={},
        copy_algebs=[],
        copy_states=[],
        service_keys=[],
        **kwargs):
    retval = True
    space4 = '    '
    space8 = space4 * 2
    """Input data consistency check"""
    to_check = {
        'param': params,
        'mandatory': mandatory,
        'zero': zeros,
        'power': powers,
        'voltage': voltages,
        'currents': currents,
        'z': z,
        'y': y,
        'dccurrent': dccurrents,
        'dcvoltage': dcvoltages,
        'r': r,
        'g': g,
        'times': times,
    }

    if not data:
        print('* Error: <data> dictionary is not defined.')
        return

    for key, val in to_check.items():
        if not val:
            continue
        for item in val:
            if item not in data.keys():
                print('* Warning: {} <{}> is not in data.'.format(key, item))

    for item in interfaces:
        if item not in algebs + copy_algebs:
            print('* Warning: interface <{}> is not defined'.format(item))

    for key, val in hard_limit.items():
        if key not in algebs:
            print('* Warning: variable <{}> in hard_limit not defined.'.format(
                key))
        for item in val:
            if type(item) in (int, float):
                pass
            elif item not in consts:
                print(
                    '* Warning: const <{}> in hard_limit not defined.'.format(
                        item))

    for key, val in windup.items():
        if key not in algebs:
            print(
                '* Warning: variable <{}> in windup not defined.'.format(key))
        for item in val:
            if type(item) in (int, float):
                continue
            elif item not in consts:
                print('* Warning: const <{}> in windup not defined.'.format(
                    item))

    for key, val in anti_windup.items():
        if key not in states:
            print(
                '* Warning: variable <{}> in anti_windup not defined.'.format(
                    key))
        for item in val:
            if type(item) in (int, float):
                continue
            elif item not in consts:
                print(
                    '* Warning: const <{}> in anti_windup not defined.'.format(
                        item))
    """Equation and variable number check"""
    nalgebs, nalgeb_eq, nstates, ndiff_eq, ninterfaces = len(algebs), len(
        algeb_eq), len(states), len(diff_eq), len(interfaces)

    if nalgebs + ninterfaces != nalgeb_eq:
        print('* Warning: there are {} algebs and {} algeb equations.'.format(
            nalgebs, nalgeb_eq))

    if nstates != ndiff_eq:
        print('* Warning: there are {} states and {} differential equations.'.
              format(nstates, ndiff_eq))

    # check for duplicate names
    var_names = consts + algebs + states + copy_algebs + copy_states
    if len(set(var_names)) != len(var_names):
        raise NameError('Duplicated names are declared!')
    # Set up sympy symbols for variables, constants and equations
    sym_consts, sym_algebs, sym_states, sym_interfaces = [], [], [], []
    sym_algebs_ext, sym_states_ext = [], []
    sym_f, sym_g, sym_serv, sym_init1 = [], [], [], []
    sym_hard_limit, sym_windup, sym_anti_windup = [], [], []

    states_anti_windup = list(anti_windup.keys())
    # algebs_windup = list(windup.keys())
    # algebs_hard_limit = list(hard_limit.keys())

    # remove interface variables in copy_algebs
    for item in interfaces:
        if item in copy_algebs:
            copy_algebs.remove(item)

    # algebs_ext = algebs + interfaces + copy_algebs
    # states_ext = states + copy_states

    for idx, var in enumerate(states):
        if var in states_anti_windup:
            tpl = '({} - {}) / {}'
            diff_eq[idx] = tpl.format(diff_eq[idx], var, anti_windup[var][0])

    # convert consts and variables into sympy.Symbol
    sym_maping = {
        'consts': sym_consts,
        'algebs': sym_algebs,
        'states': sym_states,
        'algebs_ext': sym_algebs_ext,
        'states_ext': sym_states_ext,
        'interfaces': sym_interfaces,
        'states_anti_windup': sym_anti_windup,
        'algebs_windup': sym_windup,
        'algebs_hard_limit': sym_hard_limit,
    }
    for key, val in sym_maping.items():
        for item in eval(key):
            call = '{} = Symbol(item)'.format(item)
            exec(call)
            val.append(eval(item))

    # convert service_eq.keys() into sympy.Symbols and values into equations
    # for var, eq in service_eq.items():
    if service_eq:
        for item in service_eq:
            var = item[0]
            eq = item[1]
            if var not in consts:
                print('* Warning: declaring undefined service variable <{}>'.
                      format(var))
            call = '{} = Symbol(var)'.format(var)
            exec(call)
            expr = eval('{}'.format(eq))
            sym_serv.append([eval(var), expr])

    if init1_eq:
        for item in init1_eq:
            var = item[0]
            eq = item[1]
            if var not in states + algebs:
                print('* Warning: initializing undefined variable <{}>'.format(
                    var))
            call = '{} = Symbol(var)'.format(var)
            exec(call)
            expr = eval('{}'.format(eq))
            sym_init1.append([eval(var), expr])

    # convert equations into symbolic expression
    for item in algeb_eq:
        expr = eval('{}'.format(item))
        sym_g.append(expr)

    for item in diff_eq:
        expr = eval('{}'.format(item))
        sym_f.append(expr)
    """Derive the jacobians of equation f and g.
    Save to Fx, Fy and Gx Gy in a list of three elements: [equation_idx, var_idx, expression]"""
    Fx, Fy, Gx, Gy = list(), list(), list(), list()
    for eq_idx, expr in enumerate(sym_g):
        try:
            free_syms = expr.free_symbols
        except AttributeError:
            free_syms = []

        for sym in free_syms:
            if sym in sym_algebs_ext:
                # take the derivative and go to Gy
                sym_idx = sym_algebs_ext.index(sym)
                Gy.append([eq_idx, sym_idx, expr.diff(sym)])
            elif sym in sym_states_ext:
                # take the derivative and go to Gx
                sym_idx = sym_states_ext.index(sym)
                Gx.append([eq_idx, sym_idx, expr.diff(sym)])
            else:
                pass  # skip constants

    for eq_idx, expr in enumerate(sym_f):
        try:
            free_syms = expr.free_symbols
        except AttributeError:
            free_syms = []

        for sym in free_syms:
            if sym in sym_algebs_ext:
                sym_idx = sym_algebs_ext.index(sym)
                Fy.append([eq_idx, sym_idx, expr.diff(sym)])
            elif sym in sym_states_ext:
                sym_idx = sym_states_ext.index(sym)
                Fx.append([eq_idx, sym_idx, expr.diff(sym)])
    # Save equations into callable CVXOPT functions
    fcall, gcall = [], []
    gycall, fxcall, jac0 = [], [], []
    jac0_line = 'dae.add_jac({}0, {}, self.{}, self.{})'
    call_line = 'dae.add_jac({}, {}, self.{}, self.{})'

    # format f and g equations
    fcall_anti_windup = 'dae.anti_windup(self.{0}, {1}, {2})'

    for sym, eq in zip(sym_states, sym_f):
        string_eq = stringfy(eq, sym_consts, sym_states_ext, sym_algebs_ext)
        # handling anti_windup
        template = 'dae.f[self.{0}] = {1}'
        fcall.append(template.format(sym, string_eq))
        if sym in sym_anti_windup:
            val = eval('anti_windup[\'{}\']'.format(sym))
            if type(val[1]) not in (int, float):
                val[1] = 'self.' + val[1]
            if type(val[2]) not in (int, float):
                val[2] = 'self.' + val[2]
            fcall.append(fcall_anti_windup.format(sym, val[1], val[2]))

    gcall_windup = 'dae.windup(self.{0}, self.{1}, self.{2})'
    gcall_hard_limit = 'dae.hard_limit(self.{0}, {1}, {2})'

    for sym, eq in zip(sym_algebs + sym_interfaces, sym_g):
        string_eq = stringfy(eq, sym_consts, sym_states_ext, sym_algebs_ext)
        if sym in sym_interfaces:
            template = 'dae.g += spmatrix({1}, self.{0}, [0]*self.n, (dae.m, 1), \'d\')'
        else:
            template = 'dae.g[self.{0}] = {1}'
        gcall.append(template.format(sym, string_eq))

        if sym in sym_windup:
            val = eval('windup[\'{}\']'.format(sym))
            gcall.append(gcall_windup.format(sym, val[0], val[1]))
        elif sym in sym_hard_limit:
            val = eval('hard_limit[\'{}\']'.format(sym))
            val_formatted = list(val)
            for idx, item in enumerate(val):
                if type(item) in (int, float):
                    pass
                else:
                    val_formatted[idx] = 'self.{}'.format(item)
            gcall.append(
                gcall_hard_limit.format(sym, val_formatted[0],
                                        val_formatted[1]))

    # format Jacobians
    jacobians = ['Gy', 'Gx', 'Fx', 'Fy']
    mapping = dict(
        F=sym_states_ext, G=sym_algebs_ext, y=sym_algebs_ext, x=sym_states_ext)

    for jac in jacobians:
        for item in eval(jac):
            eqname = mapping[jac[0]][item[0]]
            varname = mapping[jac[1]][item[1]]
            equation = item[2]
            try:
                free_syms = equation.free_symbols
            except AttributeError:
                free_syms = []

            string_eq = stringfy(equation, sym_consts, sym_states_ext,
                                 sym_algebs_ext)

            isjac0 = 1
            for sym in free_syms:
                if sym in sym_consts:
                    continue
                elif sym in sym_algebs_ext + sym_states_ext:
                    isjac0 = 0
                    break
                else:
                    raise KeyError

            if isjac0:
                jac0.append(jac0_line.format(jac, string_eq, eqname, varname))
            else:
                if jac == 'Gy':
                    gycall.append(
                        call_line.format(jac, string_eq, eqname, varname))
                else:
                    fxcall.append(
                        call_line.format(jac, string_eq, eqname, varname))
    """Format initialization and service calls"""
    init1call = []
    for item in sym_init1:
        rhs = stringfy(item[1], sym_consts, sym_states_ext, sym_algebs_ext)
        if item[0] in sym_algebs_ext:
            out = 'dae.y[self.{}] = {}'.format(item[0], rhs)
        elif item[0] in sym_states_ext:
            out = 'dae.x[self.{}] = {}'.format(item[0], rhs)
        else:
            raise KeyError
        init1call.append(out)

    servcall = []
    for item in ctrl:
        key, val = item
        out = 'self.copy_data_ext(\'{}\', \'{}\', \'dest={}\', idx=self.{})'.format(
            val[0], val[1], key, val[2])
        servcall.append(out)
    for item in sym_serv:
        rhs = stringfy(item[1], sym_consts, sym_states_ext, sym_algebs_ext)
        out = 'self.{} = {}'.format(item[0], rhs)
        servcall.append(out)

    calls = {
        'gcall': not not gcall,
        'fcall': not not fcall,
        'gycall': not not gycall,
        'fxcall': not not fxcall,
        'jac0': (not not jac0) or (not not sym_algebs),
        'init1': (not not init1call) or (not not servcall),
    }
    """Build function call strings"""
    out_calls = []
    if servcall:
        out_calls.append(space4 + 'def servcall(self, dae):')
        for item in servcall:
            out_calls.append(space8 + item)
        out_calls.append('')

    if init1call or servcall:
        out_calls.append(space4 + 'def init1(self, dae):')
        out_calls.append(space8 + 'self.servcall(dae)')
        for item in init1call:
            out_calls.append(space8 + item)
        out_calls.append('')

    if gcall:
        out_calls.append(space4 + 'def gcall(self, dae):')
        for item in gcall:
            out_calls.append(space8 + item)
        out_calls.append('')

    if fcall:
        out_calls.append(space4 + 'def fcall(self, dae):')
        for item in fcall:
            out_calls.append(space8 + item)
        out_calls.append('')

    if gycall:
        out_calls.append(space4 + 'def gycall(self, dae):')
        for item in gycall:
            out_calls.append(space8 + item)
        out_calls.append('')
    if fxcall:
        out_calls.append(space4 + 'def fxcall(self, dae):')
        for item in fxcall:
            out_calls.append(space8 + item)
        out_calls.append('')

    tinyGy = 'dae.add_jac(Gy0, 1e-6, self.{0}, self.{0})'
    if jac0 or sym_states:
        out_calls.append(space4 + 'def jac0(self, dae):')
        for item in jac0:
            out_calls.append(space8 + item)
        for item in sym_algebs:
            out_calls.append(space8 + tinyGy.format(item))
    """Class definitions in out_init"""
    # bulk update or extend of dict and list
    param_assign = space8 + 'self._{} = {}'
    list_extend = space8 + 'self._{}.extend({})'
    dict_update = space8 + 'self._{}.update(\n{})'

    out_init = list()  # def __init__ call strings
    out_init.append('from cvxopt import matrix, spmatrix')
    out_init.append('from cvxopt import mul, div, sin, cos, exp')
    out_init.append('from ..consts import *')
    out_init.append('from .base import ModelBase\n\n')
    out_init.append('class {}(ModelBase):'.format(name))
    if doc_string:
        out_init.append(space4 + "\"\"\"{}\"\"\"".format(doc_string))
    out_init.append(space4 + 'def __init__(self, system, name):')
    out_init.append(space8 + 'super().__init__(system, name)')
    if not group:
        print('*Error: Group name is not defined!')
    else:
        out_init.append(param_assign.format('group', add_quotes(group)))
    if name:
        out_init.append(param_assign.format('name', add_quotes(name)))

    meta_dict_upd = {
        'data': data,
        'units': units,
        'config_descr': descr,
        'ac': ac,
        'dc': dc,
        # 'ctrl': ctrl,
    }
    meta_list_ext = {
        'params': params,
        'algebs': algebs,
        'states': states,
        'fnamex': fnamex,
        'fnamey': fnamey,
        'service': service_keys,
        'mandatory': mandatory,
        'zeros': zeros,
        'powers': powers,
        'voltages': voltages,
        'currents': currents,
        'z': z,
        'y': y,
        'dccurrents': dccurrents,
        'dcvoltages': dcvoltages,
        'r': r,
        'g': g,
        'times': times,
    }

    for key in sorted(meta_list_ext.keys()):
        val = meta_list_ext[key]
        if val:
            out_init.append(list_extend.format(key, val))
    for key in sorted(meta_dict_upd.keys()):
        val = meta_dict_upd[key]
        if val:
            out_init.append(
                dict_update.format(key, pprint.pformat(val, indent=12)))

    out_init.append(space8 + 'self.calls.update({})'.format(calls))

    out_init.append(space8 + 'self._init()')
    out_init.append('')

    # write to file
    try:
        fid = open(outfile, 'w')
        for line in out_init:
            fid.write(line + '\n')
        for line in out_calls:
            fid.write(line + '\n')
        fid.close()
    except IOError:
        logger.error('IOError while writing card output.')
        retval = False

    if retval:
        logger.info(
            'Card file successfully saved to <{}> with'.format(outfile))
        logger.info(
            '* constants: {}, algebs: {}, interfaces: {}, states: {}'.format(
                len(sym_consts), len(sym_algebs), len(interfaces),
                len(sym_states)))
        logger.info('* diff equations: {}, algeb equations: {}'.format(
            len(fcall), len(gcall)))
        logger.info('* fxcall: {}, gycall: {}, jac0: {}'.format(
            len(fxcall), len(gycall), len(jac0)))

    sys.exit(0)
    # return retval


def stringfy(expr, sym_const=None, sym_states=None, sym_algebs=None):
    """Convert the right-hand-side of an equation into CVXOPT matrix operations"""
    if not sym_const:
        sym_const = []
    if not sym_states:
        sym_states = []
    if not sym_algebs:
        sym_algebs = []
    expr_str = []
    if type(expr) in (int, float):
        return expr
    if expr.is_Atom:
        if expr in sym_const:
            expr_str = 'self.{}'.format(expr)
        elif expr in sym_states:
            expr_str = 'dae.x[self.{}]'.format(expr)
        elif expr in sym_algebs:
            expr_str = 'dae.y[self.{}]'.format(expr)
        elif expr.is_Number:
            if expr.is_Integer:
                expr_str = str(int(expr))
            else:
                expr_str = str(float(expr))
            # if expr.is_negative:
            #     expr_str = '{}'.format(expr)
            # else:
            #     expr_str = str(expr)
        else:
            raise AttributeError('Unknown free symbol <{}>'.format(expr))
    else:
        nargs = len(expr.args)
        arg_str = []
        for arg in expr.args:
            arg_str.append(stringfy(arg, sym_const, sym_states, sym_algebs))

        if expr.is_Add:
            expr_str = ''
            for idx, item in enumerate(arg_str):
                if idx == 0:
                    if len(item) > 1 and item[1] == ' ':
                        item = item[0] + item[2:]
                if idx > 0:
                    if item[0] == '-':
                        item = ' ' + item
                    else:
                        item = ' + ' + item
                expr_str += item

        elif expr.is_Mul:
            if nargs == 2 and expr.args[0].is_Integer:  # number * matrix
                if expr.args[0].is_positive:
                    expr_str = '{}*{}'.format(*arg_str)
                elif expr.args[0] == Integer('-1'):
                    expr_str = '- {}'.format(arg_str[1])
                else:  # negative but not -1
                    expr_str = '{}*{}'.format(*arg_str)
            else:  # matrix dot multiplication
                if expr.args[0] == Integer('-1'):
                    # bring '-' out of mul()
                    expr_str = ', '.join(arg_str[1:])
                    expr_str = '- mul(' + expr_str + ')'
                else:
                    expr_str = ', '.join(arg_str)
                    expr_str = 'mul(' + expr_str + ')'
        elif expr.is_Function:
            expr_str = ', '.join(arg_str)
            expr_str = str(expr.func) + '(' + expr_str + ')'
        elif expr.is_Pow:
            if arg_str[1] == '-1':
                expr_str = 'div(1, {})'.format(arg_str[0])
            else:
                expr_str = '({})**{}'.format(*arg_str)
        elif expr.is_Div:
            expr_str = ', '.join(arg_str)
            expr_str = 'div(' + expr_str + ')'
        else:
            raise NotImplementedError
    return expr_str



"""
Parser for DOME RAW format 0.1
From Book "Power System Modeling and Scripting" by Dr. Federico Milano
"""
import re
import os
from math import ceil
import logging
logger = logging.getLogger(__name__)


def testlines(fid):
    return True  # hard coded yet


def alter(data, system):
    """Alter data in dm format devices"""
    device = data[0]
    action = data[1]
    if data[2] == '*':
        data[2] = '.*'
    regex = re.compile(data[2])
    prop = data[3]
    value = float(data[4])
    if action == 'MUL':
        for item in range(system.__dict__[device].n):
            if regex.search(system.__dict__[device].name[item]):
                system.__dict__[device].__dict__[prop][item] *= value
    elif action == 'REP':
        for item in range(system.__dict__[device].n):
            if regex.search(system.__dict__[device].name[item]):
                system.__dict__[device].__dict__[prop][item] = value
    elif action == 'DIV':
        if not value:
            return
        for item in range(system.__dict__[device].n):
            if regex.search(system.__dict__[device].name[item]):
                system.__dict__[device].__dict__[prop][item] /= value
    elif action == 'SUM':
        for item in range(system.__dict__[device].n):
            if regex.search(system.__dict__[device].name[item]):
                system.__dict__[device].__dict__[prop][item] += value
    elif action == 'SUB':
        for item in range(system.__dict__[device].n):
            if regex.search(system.__dict__[device].name[item]):
                system.__dict__[device].__dict__[prop][item] -= value
    elif action == 'POW':
        for item in range(system.__dict__[device].n):
            if regex.search(system.__dict__[device].name[item]):
                system.__dict__[device].__dict__[prop][item] **= value
    else:
        print('ALTER action <%s> is not defined', action)


def read(file, system, header=True):
    """Read a dm format file and elem_add to system"""
    retval = True
    fid = open(file, 'r')
    sep = re.compile(r'\s*,\s*')
    comment = re.compile(r'^#\s*')
    equal = re.compile(r'\s*=\s*')
    math = re.compile(r'[*/+-]')
    double = re.compile(r'[+-]? *(?:\d+(?:\.\d*)?|\.\d+)(?:[eE][+-]?\d+)?')

    # parse data
    while True:
        line = fid.readline()
        if not line:
            break
        line = line.replace('\n', '')
        line = line.strip()
        if not line:
            continue
        if comment.search(line):
            continue
        # span multiple line
        while line.endswith(',') or line.endswith(';'):
            newline = fid.readline()
            line = line.replace('\n', '')
            newline = newline.strip()
            if not newline:
                continue
            if comment.search(newline):
                continue
            line += ' ' + newline
        data = sep.split(line)
        device = data.pop(0)
        device = device.strip()
        if device == 'ALTER':
            alter(data, system)
            continue
        if device == 'INCLUDE':
            logger.debug('Parsing include file <{}>'.format(data[0]))
            newpath = data[0]
            if not os.path.isfile(newpath):
                newpath = os.path.join(system.files.path, data[0])
                if not os.path.isfile(newpath):
                    raise FileNotFoundError(
                        'Unable to locate file in {}'.format(newpath))
            read(newpath, system, header=False)  # recursive call
            logger.debug('Parsing of include file <{}> completed.'.format(
                data[0]))
            continue
        kwargs = {}
        for item in data:
            pair = equal.split(item)
            key = pair[0].strip()
            value = pair[1].strip()
            if value.startswith('"'):
                value = value[1:-1]
            elif value.startswith('['):
                array = value[1:-1].split(';')
                if math.search(value):  # execute simple operations
                    value = list(map(lambda x: eval(x), array))
                else:
                    value = list(map(lambda x: float(x), array))
            elif double.search(value):
                if math.search(value):  # execute simple operations
                    value = eval(value)
                else:
                    value = float(value)
            elif value == 'True':
                value = True
            elif value == 'False':
                value = False
            else:
                value = int(value)
            kwargs[key] = value
        index = kwargs.pop('idx', None)
        namex = kwargs.pop('name', None)
        try:
            system.__dict__[device].elem_add(idx=index, name=namex, **kwargs)
        except KeyError:
            logger.error(
                'Error adding device {:s} to powersystem object.'.format(
                    device))
            logger.debug(
                'Make sure you have added the jit models in __init__.py'
            )

    fid.close()
    return retval


def write(file, system):
    """
    Write data in system to a dm file
    """

    # TODO: Check for bugs!!!

    out = list()
    out.append('# DOME format version 1.0')
    ppl = 7  # parameter per line
    retval = True
    dev_list = sorted(system.devman.devices)
    for dev in dev_list:
        model = system.__dict__[dev]
        if not model.n:
            continue

        out.append('')
        header = dev + ', '
        space = ' ' * (len(dev) + 2)
        keys = list(model._data.keys())
        keys.extend(['name', 'idx'])
        keys = sorted(keys)

        # remove non-existent keys
        for key in keys:
            if key not in model.__dict__.keys():
                keys.pop(key)

        nline = int(ceil(len(keys) / ppl))
        nelement = model.n
        vals = [''] * len(keys)

        # for each element, read values
        for elem in range(nelement):
            for idx, key in enumerate(keys):
                if model._flags['sysbase'] and key in model._store.keys():
                    val = model._store[key][elem]
                else:
                    val = model.__dict__[key][elem]

                if isinstance(val, float):
                    val = round(val, 5)
                elif isinstance(val, str):
                    val = '"{}"'.format(val)
                elif isinstance(val, list):
                    val = list(val)
                    val = '; '.join(str(i) for i in val)
                    val = '[{}]'.format(val)
                elif val is None:
                    val = 0
                vals[idx] = val

            pair = []
            for key, val in zip(keys, vals):
                pair.append('{} = {}'.format(key, val))

            for line in range(nline):
                string = ', '.join(pair[ppl * line:ppl * (line + 1)])
                if line == 0:  # append header or space
                    string = header + string
                else:
                    string = space + string
                if not line == nline - 1:  # add comma except for last line
                    string += ','

                out.append(string)

    fid = open(file, 'w')
    for line in out:
        fid.write(line + '\n')

    fid.close()
    return retval
